// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Nicolas Pfeiffer, Florian Schmaus
#include "Fibril.hpp"

thread_local Fibril* Fibril::toResume = nullptr;

void Fibril::run() {
	Context* currentContext = Context::getCurrentContext();
	LOGD("run() of " << *this << " on context " << currentContext);

	if constexpr (LOCKED_CONTINUATION_STEALING) {
		if (!activeChildrenCount)
			activeChildrenCount = 2;
		else
			activeChildrenCount++;

		fibrilMutex.unlock();
	} else {
		reverseStealCount -= 1;
	}

	// TODO: Fibril stats: increase steal count by one here.

	// This is the run() method that is executed on a fresh context. We
	// must ensure that this context is in all cases free'd again, via
	// putFreeContext(). This run() method must eventually call
	// Fibril::join(), i.e. perform an explicit join operation. Since we
	// are in the run() method, a cactus stack was formed and join()
	// will always call ContextManager::tryResumeFibril(), which leads
	// to ContextManager::start() where we free (at some point in the
	// future hopefully re-use) this context in case this Fibril could
	// not yet be resumed.

	// Reserve 128 byte at the bottom for the linking region. Basically
	// this allows the compiler to pop args off the stack, but
	// unfortunately on the new stack. If we do not reserve space for
	// this, then it would lead to access outside (above) the stack
	// area. See Nicolas Pfeiffer's Masters thesis [pfeiffer2020cactus]
	// § 4.7 for more information.
	cont.execute((void**)currentContext->getTos() - 16);
}

void Fibril::printTo(std::ostream& strm, bool withPtr) const {
	// clang-format off
	strm << "Fibril [";
	if (withPtr) {
		strm << "ptr=" << this << ", ";
	}
	strm << "initialContext=" << initialContext
			 << ", " << cont
			 << "]";
	// clang-format-off
}

void emper::maybeLockFibril(void* queueItem) {
	auto* abstractFiber = static_cast<AbstractFiber*>(queueItem);
	Fibril* fibril = abstractFiber->asFibrilIfPossible();
	if (fibril) {
		fibril->fibrilMutex.lock();
	}
}
