// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020 Florian Schmaus, Florian Fischer
#pragma once

#include <stdarg.h>	 // NOLINT(modernize-deprecated-headers)
#include <stddef.h>	 // NOLINT(modernize-deprecated-headers)

#include "emper-common.h"
#include "emper-config.h"

#ifdef EMPER_IO
#include <sys/socket.h>
#endif

typedef struct runtime runtime;	 // NOLINT(modernize-use-using)

typedef struct fiber fiber;	 // NOLINT(modernize-use-using)

typedef struct bps bps;	 // NOLINT(modernize-use-using)

typedef struct cps cps;	 // NOLINT(modernize-use-using)

#ifdef __cplusplus
extern "C" {
#endif

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
runtime* init_runtime(void);

void wait_until_runtime_finished();

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
fiber* fiber_from(void (*function)(void*), void* arg);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
fiber* fiber_from0(void (*function)());

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
fiber* aff_fiber_from(void (*function)(void*), void* arg, workeraffinity_t* affinity);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
fiber* aff_fiber_from0(void (*function)(), workeraffinity_t* affinity);

void init_affinity(workeraffinity_t affinity[], unsigned int n);

void schedule(fiber* fiber);

void schedule_from_anywhere(fiber* fiber);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
bps* new_binary_sem(void);

void signal_bps(bps* sem);

void signal_and_exit_bps(bps* sem);

void wait_bps(bps* sem);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
cps* new_counting_sem(void);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
cps* new_counting_sem_with(unsigned int i);

void signal_cps(cps* sem);

void signal_and_exit_cps(cps* sem);

void wait_cps(cps* sem);

#ifdef EMPER_IO
// NOLINTNEXTLINE(modernize-use-trailing-return-type)
ssize_t emper_recv(int socket, void* buffer, size_t length, int flags);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
ssize_t emper_send(int socket, const void* buffer, size_t length, int flags);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_connect(int socket, const struct sockaddr* address, socklen_t address_len);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_accept(int socket, struct sockaddr* address, socklen_t* address_len);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
ssize_t emper_read_file(int fildes, void* buf, size_t nbyte, off_t offset);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
ssize_t emper_write_file(int fildes, const void* buf, size_t nbyte, off_t offset);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
ssize_t emper_writev(int fildes, const struct iovec* iov, int iovcnt);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_open(const char* pathname, int flags, ...);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_openat(int dirfd, const char* pathname, int flags, ...);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_shutdown(int sockfd, int how);

// NOLINTNEXTLINE(modernize-use-trailing-return-type)
int emper_close(int fd);
#endif

#ifdef __cplusplus
}
#endif
