// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Fischer
#pragma once

#include <pthread.h>	// for pthread_t

#include <cstdint>
#include <string>

#include "emper-common.h"
#include "io/IoContext.hpp"
class Runtime;

namespace emper::io {
class GlobalIoContext : public IoContext {
	friend Runtime;
	friend void emper::log::log(const std::string& prefix, const std::string& message);
	friend class Future;
	friend class SendFuture;
	friend class RecvFuture;

 private:
	GlobalIoContext(Runtime& runtime, workerid_t workerCount, unsigned unboundedIowMax)
			: IoContext(runtime, workerCount, unboundedIowMax) {
		worker = nullptr;
	}
	GlobalIoContext(Runtime& runtime, workerid_t workerCount)
			: GlobalIoContext(runtime, workerCount, IoContext::getDefaultUnboundedIowMax()) {}

	enum class PointerTags : uint16_t { Future, Callback, IoContext, TerminationEvent };

	// pthread used to monitor the CQs from worker io_urings
	// as well as handling IO requests submitted from anywhere
	pthread_t completer;

	/* Loop executed by the completer thread */
	void completerLoop();

	// start the completer thread
	void startCompleter();

	void initiateTermination();
	void waitUntilFinished() const;

	void prepareWorkerNotification(IoContext& workerIo);
};
}	 // namespace emper::io
