// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Fischer, Florian Schmaus
#pragma once

#include <string>

class TestLinuxVersion;

namespace emper::lib {
class LinuxVersion {
	friend class ::TestLinuxVersion;
	static auto getUtsRelease() -> const std::string&;
	std::string version;

	auto compare(const std::string& s) -> int;

	LinuxVersion(std::string version) : version(version) {}

 public:
	LinuxVersion();

	auto operator>(const std::string& s) -> bool { return compare(s) > 0; }

	auto operator<(const std::string& s) -> bool { return compare(s) < 0; }

	auto operator==(const std::string& s) -> bool { return compare(s) == 0; }

	auto operator>=(const std::string& s) -> bool { return compare(s) >= 0; }

	auto operator<=(const std::string& s) -> bool { return compare(s) <= 0; }
};
}	 // namespace emper::lib

#define EMPER_LINUX_EQ(version) (emper::lib::LinuxVersion() == version)
#define EMPER_LINUX_GT(version) (emper::lib::LinuxVersion() > version))
#define EMPER_LINUX_GE(version) (emper::lib::LinuxVersion() >= version)
#define EMPER_LINUX_LT(version) (emper::lib::LinuxVersion() < version)
#define EMPER_LINUX_LE(version) (emper::lib::LinuxVersion() <= version)
