// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2022 Florian Schmaus
#pragma once

#include <cstddef>
#include <utility>

#include "Common.hpp"

namespace emper::lib::template_util {

// https://stackoverflow.com/a/66043076/194894
template <typename T>
auto getSize(T& t) -> size_t {
	typename T::size_type size = t.size();
	size_t value_type_size = sizeof(typename T::value_type);

	return size * value_type_size;
}

template <typename T, std::enable_if_t<std::is_integral<T>::value, bool> = true>
std::pair<T, std::string> toHumanReadableBytes(T bytes) {
	// TODO: Ideally the std::string would be const.
	const static std::vector<std::string> SI_BYTES = {"B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB"};

	for (auto unit : SI_BYTES) {
		if (bytes < 1024) return std::make_pair(bytes, unit);

		bytes >>= 10;
	}

	DIE;
}

}	 // namespace emper::lib::template_util
