// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2021-2022 Florian Schmaus
#pragma once

#include <cstddef>	// for size_t
#include <optional>

#include "Fiber.hpp"
#include "Scheduler.hpp"
#include "emper-common.h"
#include "lib/adt/MpscQueue.hpp"
#include "lib/adt/WsQueue.hpp"

struct NextFiberResult;
class AbstractFiber;
class Runtime;
class RuntimeStrategy;

class AbstractWorkStealingScheduler : public Scheduler {
	template <size_t SIZE>
	using WsQueue = lib::adt::wsqueue::Scheduler<AbstractFiber*, SIZE>;
	using MpscQueue = adt::MpscQueue<Fiber>;

 public:
	static const int QUEUE_SIZE = 1024;

 private:
	auto nextFiberViaAnywhereQueue() -> std::optional<NextFiberResult>;
	auto tryStealFiberFrom(workerid_t victim) -> std::optional<NextFiberResult>;

	bool queueFullWarningEmitted = false;

 protected:
	WsQueue<QUEUE_SIZE>** queues;
	static thread_local WsQueue<QUEUE_SIZE> queue;

	MpscQueue** mpscQueues;
	static thread_local MpscQueue mpscQueue;

	void scheduleViaWorkStealing(Fiber& fiber);
	void scheduleToMpscQueue(Fiber& fiber, workerid_t workerId);

	auto maybeRecycle(AbstractFiber* fiber) -> bool;

	// This method is static because it only uses the thread_local mpscQueue
	static auto nextFiberResultFromMpscQueue() -> std::optional<NextFiberResult>;
	auto nextFiberResultViaWorkStealing() -> std::optional<NextFiberResult>;
	auto nextFiberResultFromMpscQueueOrWorkStealing() -> std::optional<NextFiberResult>;

	auto pushBottom(AbstractFiber& fiber) -> bool override;

	auto popBottom() -> AbstractFiber* override {
		AbstractFiber* fiber;
		if (queue.popBottom(&fiber)) {
			return fiber;
		}
		return nullptr;
	}

 public:
	AbstractWorkStealingScheduler(Runtime& runtime, RuntimeStrategy& strategy);

	friend class Runtime;
};
