// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus, Florian Fischer
#pragma once

#include <optional>

#include "CallerEnvironment.hpp"
#include "Fiber.hpp"
#include "emper-common.h"
#include "lib/adt/MpscQueue.hpp"
#include "strategies/AbstractWorkStealingScheduler.hpp"

class Runtime;
class RuntimeStrategy;
struct NextFiberResult;

class LawsScheduler : public AbstractWorkStealingScheduler {
	using LawsMpscQueue = adt::MpscQueue<Fiber>;

 private:
	LawsMpscQueue**& priorityQueues = AbstractWorkStealingScheduler::mpscQueues;

	template <CallerEnvironment callerEnvironment>
	void tryScheduleToPriorityQueue(Fiber& fiber);

 protected:
	void scheduleInternal(Fiber& fiber) override;
	void scheduleOnInternal(Fiber& fiber, workerid_t workerId) override {
		scheduleToMpscQueue(fiber, workerId);
	}
	void scheduleFromAnywhereInternal(Fiber& fiber) override;
	void scheduleFromAnywhereInternal(Fiber** fibers, unsigned count) override;

 public:
	LawsScheduler(Runtime& runtime, RuntimeStrategy& strategy);

	auto nextFiber() -> std::optional<NextFiberResult> override;
};
