// SPDX-License-Identifier: LGPL-3.0-or-later
// Copyright © 2020-2021 Florian Schmaus
#pragma once

#include <functional>
#include <memory>

#include "strategies/AbstractWorkStealingStrategy.hpp"
#include "strategies/ws/WsDispatcher.hpp"
#include "strategies/ws/WsScheduler.hpp"

class Runtime;
class RuntimeStrategyStats;
class WsStrategyFactory;
class WsStrategyStats;

class WsStrategy final : public AbstractWorkStealingStrategy {
 private:
	WsScheduler scheduler;
	WsDispatcher dispatcher;

	WsStrategy(Runtime& runtime);

	~WsStrategy() override = default;

	auto getScheduler() -> WsScheduler& override;

	auto getDispatcher() -> WsDispatcher& override;

 public:
	static auto getFactory() -> std::function<WsStrategy*(Runtime&)>;

	auto getStats() -> std::shared_ptr<RuntimeStrategyStats> override;

	friend WsScheduler;
	friend WsDispatcher;
	friend WsStrategyFactory;
	friend WsStrategyStats;
};
